/*   CHISQ, GFIT and GAUSS
 *   Gauss fit used in SA
 *
 *    function --gfit()--- fits up to four Gauss functions
 *    to the data.
 *
 *    j. m. xu
 */

#include <math.h>
#include "dacf.h"

#define  D1  0.075
#define  D2  0.1
#define  D3  0.05

extern  float *pdat;
extern  int  npks;		/* number of peaks to be fitted */
extern  int *pmarks;
extern double bgd(double a);
extern double matinv (double array[][12], int order);

double  gpar[12], sigm[12], chi;
double  delsav[12], deriv[12];
int     npar, nfree;


double  gauss (double xi, double par[])		/* value of gauss function */
{
    double  peak = 0.0, x, sigma, height;
    register int    j, n;
    extern double  bgd (double a);

    for (j = 0; j < npks; j++) {
 	n = j * 3;
	x = xi - par[n + 1];
	sigma = par[n + 2] * par[n + 2] / 2.776;
	x = x * x / sigma;
	height = 0.939439 * par[n] / par[n + 2];
	if (x < 10.)
	  peak += exp (-x) * height;
    }
    return (peak + bgd ((double) (xi)));
}

void gderiv (double i)			/* partial derivatives */
{
    register int    j;
    double  delta, yfit1, yfit2, tmp;

    for (j = 0; j < npar; j++) {
	delta = delsav[j];
	tmp = gpar[j];
	gpar[j] += delta;
	yfit1 = gauss (i, gpar);
	gpar[j] = tmp - delta;
	yfit2 = gauss (i, gpar);
	deriv[j] = (yfit1 - yfit2) / (2.* delta);
	gpar[j] = tmp;
    }
}

double  chisq (double par[])		/* function chi square */
{
    register int    j;
    double  gs;
    double  data, datn;

    chi = 0.0;

    if (nfree <= 0)
      return (0.0);
    for (j = pmarks[2]; j < pmarks[3]; j++) {
	data = pdat[j];
	if (pdat[j] == 0.0)
	  datn = 1.0e-06;
	else
	  datn = data;
	gs = gauss ((double) j, par);
	chi += (data - gs) * (data - gs) / datn;
    }
    return (chi / nfree);
}


void gfit () 			/* gauss fit */
{
    extern double   bp[];

    double  array[12][12];
    double  chisav, dat, bf, delp, lambda = 0.001;
    double  alpha[12][12], beta[12], pfit[12];
    double  data;
    register int    j, k, i;
    int     npts;
    int     m2, m3;
    double  hgt[4], area, hsum;	/* estimate value for peak parameters */

    m2 = pmarks[2];
    m3 = pmarks[3];

    npts = m3 - m2 + 1;

    area = 0.0;
    for (i = 0; i < 4; i++)
      hgt[i] = 0.0;
    hsum = 0.0;
    for (i = m2, k = 0; i < m3; i++) {
	data = pdat[i] - bgd ((double)(i));
	area += data;
	for (k = 0; k < npks; k++)
	  if (i == pmarks[k + 6]) {
	      hgt[k] = data;
	      hsum += hgt[k];
	  }
    }
    npar = 3 * npks;
    for (k = 0; k < npks; k++) {
	gpar[3 * k] = area * hgt[k] / hsum;             /* area of peaks   */
	gpar[3 * k + 1] = pmarks[k + 6];                /* center of peaks */
	gpar[3 * k + 2] = 0.939 * gpar[3 * k] / hgt[k];	/* max half width  */
    }

    /* 
     *	start iterative gauss fit -- essentially curfit
     *      optimize chi square value by increasing or
     *	decreasing the lambda until changes less than 1%
     */

    nfree = npts - npar;
    chi = chisq (gpar);
    for (k = 0; k < npks; k++) {
	i = k * 3;
	delsav[i] = D1 * gpar[i];
	delsav[i + 1] = D2;
	delsav[i + 2] = D3 * gpar[i + 2];
    }

    for (;;) {
	chisav = chi;
	for (i = 0; i < npar; i++) {
	    beta[i] = 0.0;
	    for (j = 0; j < npar; j++)
	      alpha[i][j] = 0.0;
	}

	for (i = m2; i < m3; i++) {
	    gderiv ((double) i);
	    if(pdat[i] == 0.0) pdat[i] = 1.0e-06;
	    dat = (pdat[i] - gauss ((double) i, gpar)) / pdat[i];

	    for (j = 0; j < npar; j++) {
		beta[j] = beta[j] + dat * deriv[j];

		for (k = 0; k <= j; k++)
		    alpha[j][k] = alpha[j][k] + (deriv[j] * deriv[k])/pdat[i];
		if (alpha[j][j] <= 0.000001)
		    alpha[j][j] = 1.0;
	    }
	}
	for (j = 0; j < npar; j++)
	    for (k = 0; k <= j; k++)
		alpha[k][j] = alpha[j][k];

	for (;;) {		/*   iteration loop   */
	    for (j = 0; j < npar; j++) {
		for (k = 0; k < npar; k++)
		    array[j][k] = alpha[j][k] / sqrt (alpha[j][j]*alpha[k][k]);
		array[j][j] = 1 + lambda;
	    }
	    matinv (array, npar);
	    for (i = 0; i < npar; i++)
	      pfit[i] = gpar[i];
	    for (j = 0; j < npar; j++)
	      for (k = 0; k < npar; k++)
		pfit[j] += beta[k] * array[j][k] /
		  sqrt(alpha[j][j] * alpha[k][k]);

	    chi = chisq (pfit);
	    if (chisav >= chi)
	      break;
	    else
	      lambda *= 10.0;
	}
	lambda /= 10.0;
	for (i = 0; i < npar; i++)
	  gpar[i] = pfit[i];
	if (chisav <= chi * 1.001)
	  break;
    }				/*    end of loop   */

    for (i = 0; i < npar; i++)
      sigm[i] = sqrt (array[i][i] / alpha[i][i]);
    for (i = 0; i < npks; i++) {
	k = 3 * i;
	bf = 1.7 * gpar[k + 2] * bp[0];
	delp = gpar[k] + bf * bf / (gpar[k] + bf);
	sigm[k] = sqrt (delp + sigm[k] * sigm[k]);
    }
}

/*    This is a special version for up/down where
 *    only the peak area gpar[0], gpar[3],.. is fitted. Position and
 *    FWHM are kept constant.
 */


void gderiv1 (double i)			/* partial derivatives */
{
    register int    j,k;
    double delta, yfit1, yfit2;
    double tmp;

    for (j = 0; j < npks; j++) {
	k = 3*j;
	delta = delsav[k];
	tmp = gpar[k];
	gpar[k] += delta;
	yfit1 = gauss (i, gpar);
	gpar[k] = tmp - delta;
	yfit2 = gauss (i, gpar);
	deriv[k] = (yfit1 - yfit2) / (2.* delta);
	gpar[k] = tmp;
    }
}

void gfit1 () 			/* gauss fit */
{
    extern double   bp[];

    double  array[12][12];
    double  chisav, dat, bf, delp, lambda = 0.001;
    double  alpha[12][12], beta[12], pfit[12];
    double  data;

    register int    j, k, i;
    int     npts;
    int     m2, m3;

    m2 = pmarks[2];
    m3 = pmarks[3];
    npts = m3 - m2 + 1;
    nfree = npts - npks;	/* only the peakarea is fitted */
    gpar[0] = 0.0;

    for (i = m2, k = 0; i < m3; i++) {
	data = pdat[i] - bgd ((double)i);
	gpar[0] += data;
    }
    chi = chisq (gpar);
    for (k = 0; k < npks; k++) {
	i = k * 3;
	delsav[i] = D1 * gpar[i];
    }
    for (i = 0; i < npar; i++)
      deriv[i] = 0.0;

    for (;;) {
	chisav = chi;
	for (i = 0; i < npar; i++) {
	    beta[i] = 0.0;
	    for (j = 0; j < npar; j++)
		alpha[i][j] = 0.0;
	}

	for (i = m2; i < m3; i++) {
	    gderiv1 ((double) i);
	    dat = (pdat[i] - gauss ((double) i, gpar)) / pdat[i];

	    for (j = 0; j < npar; j++) {
		beta[j] = beta[j] + dat * deriv[j];

		for (k = 0; k <= j; k++)
		    alpha[j][k] = alpha[j][k] + (deriv[j] * deriv[k])/pdat[i];
		if (alpha[j][j] <= 0.000001)
		    alpha[j][j] = 1.0;
	    }
	}
	for (j = 0; j < npar; j++)
	    for (k = 0; k <= j; k++)
		alpha[k][j] = alpha[j][k];

	for (;;) {		/*   iteration loop   */
	    for (j = 0; j < npar; j++) {
		for (k = 0; k < npar; k++)
		    array[j][k] = alpha[j][k] / sqrt (alpha[j][j]*alpha[k][k]);
		array[j][j] = 1 + lambda;
	    }
	    matinv (array, npar);
	    for (i = 0; i < npar; i++)
		pfit[i] = gpar[i];
	    for (j = 0; j < npar; j++)
		for (k = 0; k < npar; k++)
		    pfit[j] += beta[k] * array[j][k] /
		      sqrt(alpha[j][j] * alpha[k][k]);

	    chi = chisq (pfit);
	    if (chisav >= chi)
		break;
	    else
		lambda *= 10.0;
	}
	lambda /= 10.0;
	for (i = 0; i < npar; i++)
	    gpar[i] = pfit[i];
	if (chisav <= chi * 1.001)
	    break;
    }				/*    end of loop   */

    for (i = 0; i < npar; i++)
	sigm[i] = sqrt (array[i][i] / alpha[i][i]);
    for (i = 0; i < npks; i++) {
	k = 3 * i;
	bf = 1.7 * gpar[k + 2] * bp[0];
	delp = gpar[k] + bf * bf / (gpar[k] + bf);
	sigm[k] = sqrt (delp + sigm[k] * sigm[k]);
    }
}

